﻿/*=======================================================================

  Program:   
  Module:    $RCSfile: $
  Language:  C++
  Date:      $Date: $
  Version:   $Revision: $
  Author:    $Author: $
  Comment:   

  Vicomtech-IK4 - visual interaction & communications technologies
  http://www.vicomtech.org

=========================================================================*/


#ifndef ImageToPolyDataFilter_h
#define ImageToPolyDataFilter_h


// ITK includes
#include "itkImage.h"
#include "itkImageFileReader.h"
#include "itkImageToVTKImageFilter.h"
#include "itkProcessObject.h"

// VTK class declarations
class vtkPolyData;
class vtkContourFilter;
class vtkPolyDataNormals;
class vtkImageGaussianSmooth;
class vtkImageThreshold;
class vtkDecimatePro;


/** \class ImageToPolyDataFilter
 *  \brief Generates polygonal 3D mesh from an input 3D image.
 *
 * This class generates a polygonal mesh (surface) from a 3D input image in ITK format. The 3D surface
 * is obtained by iso-contouring the 3D image with a given an iso-value. Typically the input to this
 * filter is a 3D mask obtained from a segmentation but a full grayscale image can be provided.
 *
 * Several options are provided to smooth the resulting mesh in two ways:
 * - By smoothing the input image prior to 3D reconstruction using a Gaussian filter.
 * - By smoothing polygons after reconstruction.
 *
 * An option is provided to decrease the number of polygons of the resulting surface.
 *
 * DESCRIPTION OF OPERATIONS
 *
 * - Set input 3D image in ITK format.
 * - Set/Get if the input image is a segmentation mask or a full image.
 * - Get output 3D mesh.
 * - Update/calculate 3D mesh in same/separate thread.
 * - Set/Get Iso-value.
 * - Enable/Disable Gaussian smoothing.
 * - Set/Get sigma of Gaussian kernel for smoothing.
 * - Set/Get maximum feature angle for image smoothing.
 * - Set/Get maximum number of polygons in polygon reduction.
 * - Set/Get decimation target for polygon reduction.
 * 
 * DATA ACCESS
 *
 * - This class accesses image data without modifying it.
 *
 * ERROR HANDLING
 *
 * - This class throws no exceptions.
 *
 * SEE ALSO
 *
 * - ProcessObject
 * - SegmentationManager
 * 
 * 
 */


enum PolyDataSmoothingStrenghType
{
  ZERO_SMOOTHING,
  MIN_SMOOTHING,
  LOW_SMOOTHING,
  MEDIUM_SMOOTHING,
  HIGH_SMOOTHING
};


template <class TImage>
class ImageToPolyDataFilter : public itk::ProcessObject
{
  
public:

  /** Standard class typedefs. */
  typedef ImageToPolyDataFilter         Self;
  typedef ProcessObject                 Superclass;

  typedef itk::SmartPointer<Self>       Pointer;
  typedef itk::SmartPointer<const Self> ConstPointer;

  typedef TImage                        ImageType;
  typedef typename ImageType::Pointer   ImagePointer;

  /** itk to vtk typedefs */
  typedef itk::ImageToVTKImageFilter< TImage >              ITKToVTKImageAdapterType;
  typedef typename ITKToVTKImageAdapterType::Pointer        ITKToVTKImageAdapterPointer;

public:

  /** Standard macro for new objects.  */
  itkNewMacro( Self );
 //static ImageToPolyDataFilter* New();

  /** Standard type macro. */
  itkTypeMacro( ImageToPolyDataFilter, ProcessObject );

  /** Set the input image. */
  virtual void SetInput( TImage *image );

  /** Get the output VTK poly data. */
  vtkPolyData * GetOutput();

  /* return the VtkImage from the ITK to VTK Adapter */
  vtkDataSet * GetVtkImage();
  
  /** Update this ProcessObject using a worker thread. By default this creates a ThreadProcessor 
    * passing a pointer to this object and calling run. Reimplemented from ProcessObject to connect
    * the internal pipeline first. */
  virtual void ThreadedUpdate();

  /** Update the pipeline. Users of the class should call ThreadedUpdate() instead
    * to use multi-threading capabilities. */
  virtual void Update();

  /** Set/Get the contour iso-value. When we are generating the surface from the original image this
    * is a real iso-value in HU. If we are generating from the mask this is the label
    * value of the segmented object. */
  itkSetMacro( IsoValue, double );
  itkGetConstMacro( IsoValue, double );

  /** Set/Get the number of smoothing iterations. */
  //void SetNumberOfWidowedSincSmoothingIterations( unsigned int iter );
  //  itkGetMacro( NumberOfWindowedSincSmoothingIterations, unsigned int );

  /** Methods to set the number of iterations of the Widowed Sinc Filter*/
  //void SetSmoothingStrengh(PolyDataSmoothingStrenghType strengh);
  
  /** Enable Disable the gaussian image smoothing */
  itkSetMacro( SmoothingSigma, double );
  itkGetConstMacro( SmoothingSigma, double );

  /** Feature angle for polygon reduction. */
  itkSetMacro( FeatureAngle, double );
  itkGetConstMacro( FeatureAngle, double );

  /** Set/Get the maximum number of polygons to render. If the final number exceeds this 
  * number an appropiate decimation target is set. If 0 all polygons are rendered. 
  * This is the default. */
  itkSetMacro( MaxPolygons, unsigned long );
  itkGetConstMacro( MaxPolygons, unsigned long );

  /** Enable Disable the gaussian image smoothing */
  void EnableGaussianSmoothing( double sigma = 1.5 )
    { m_SmoothingSigma = sigma; }

  void DisableGaussianSmoothing()
    { m_SmoothingSigma = 0; }
    
  /** Check if gaussian smoothing is enabled. This is true when sigma != 0. */
  bool IsGaussianSmoothingEnabled()
    { if( m_SmoothingSigma ) return true; else return false; }

  /** Set/Get the flag that controls if we are using the source CT image (false)
    * or the segmented image (true). */
  itkGetMacro( InputIsMaskImage, bool );
  itkSetMacro( InputIsMaskImage, bool );
  itkBooleanMacro( InputIsMaskImage );
  
  /** Set the target value for PolyData decimation. Target ranges from 0.0 to 1.0 and
    * sets the percentage of decimation (0.3 = 30% of reduction). */
  void SetDecimationTarget( float target );

protected:

  /** Default constructor. */
  ImageToPolyDataFilter();

  /** Default destructor. */
  virtual ~ImageToPolyDataFilter();

  /** Make the pipeline connections. */
  void ConnectPipeline();

  /** Print support. */
  //virtual void PrintSelf(std::ostream& os, Indent indent) const;

private:

  ImageToPolyDataFilter( const Self & ); // purposely not implemented
  void operator =( const Self & ); // purposely not implemented

protected:

  /** Input image. */
  ImagePointer                 m_Input;
   
  /** VTK to ITK image adaptor Smart Pointer for the output. */
  ITKToVTKImageAdapterPointer  m_ITKToVTKAdapter;

  /** Filter that extracts a surface from a Meta Image. */
  vtkContourFilter            *m_ContourFilter;
  
  /** filter that generates the normals of a surface */
  vtkPolyDataNormals          *m_PolyDataNormals;

  /** Windowed Sinc Filter. */
  //vtkWindowedSincPolyDataFilter   * m_WindowedSincPolyDataFilter;

  /** Gaussian Filter for smoothing */
  vtkImageGaussianSmooth      *m_GaussianFilter;

  /** Filter to threshold values in case we are selecting labels from the mask. */
  vtkImageThreshold           *m_ImageThreshold;

  /** Decimation filter */
  vtkDecimatePro              *m_DecimatePro;

  /** number of iteration of the smoothing filter */
  //unsigned int   m_NumberOfWindowedSincSmoothingIterations;

  /** Standard deviation for the Gaussian smoothing. */
  double                       m_SmoothingSigma;

  /** Current iso-value. When we are generating the surface from the original image this
    * is a real iso-value in HU. If we are generating from the mask this is the label
    * value of the segmented object. */
  double                       m_IsoValue;

  /** Flag that controls if we are segmenting from the original image (false) or from
    * the segmented image (true). */
  bool                         m_InputIsMaskImage;

  /** Feature angle for polygon reduction. */
  double                       m_FeatureAngle;

  /** Maximum number of polygons to render. If 0 all polygons are rendered. */
  unsigned long                m_MaxPolygons;
};


#include "ImageToPolyDataFilter.txx"


#endif // ImageToPolyDataFilter_h