﻿/*=======================================================================

  Program:   
  Module:    $RCSfile: $
  Language:  C++
  Date:      $Date: $
  Version:   $Revision: $
  Author:    $Author: $
  Comment:   

  Vicomtech-IK4 - visual interaction & communications technologies
  http://www.vicomtech.org

=========================================================================*/


#ifndef ImageToPolyDataFilter_txx
#define ImageToPolyDataFilter_txx


// Project-specific includes
#include "ImageToPolyDataFilter.h"

// VTK includes
#include "vtkContourFilter.h"
#include "vtkDecimatePro.h"
#include "vtkImageGaussianSmooth.h"
#include "vtkImageThreshold.h"
#include "vtkMetaImageWriter.h"
#include "vtkPolyData.h"
#include "vtkPolyDataNormals.h"
//#include "vtkWindowedSincPolydataFilter.h"



/*----------------------------------------------------------------------*/
template <class TImage>
ImageToPolyDataFilter<TImage>::ImageToPolyDataFilter() :
  m_ContourFilter(0),
  m_PolyDataNormals(0),
  m_DecimatePro(0),
  //m_WindowedSincPolyDataFilter(0),
  //m_NumberOfWindowedSincSmoothingIterations(0),
  m_GaussianFilter(0),
  m_SmoothingSigma(1.5),
  m_ImageThreshold(0),
  m_IsoValue(127.5),
  m_InputIsMaskImage(false),
  m_FeatureAngle(25),
  m_MaxPolygons(0)
{
  // initial pipeline 
  m_ITKToVTKAdapter = ITKToVTKImageAdapterType::New();

  m_DecimatePro = vtkDecimatePro::New();
  m_DecimatePro->ReleaseDataFlagOn();
  m_DecimatePro->PreserveTopologyOn();
  m_DecimatePro->SetTargetReduction(0.0); //no decimation by default, this has no overhead
  
  m_ContourFilter   = vtkContourFilter::New();
  m_ContourFilter->ReleaseDataFlagOn();
  m_ContourFilter->SetValue(0,m_IsoValue);  //set the default iso value to 127.5, the input data is a short type image 
                                            //so 127.5 correspond to the middle value

  m_PolyDataNormals = vtkPolyDataNormals::New();
  m_PolyDataNormals->ReleaseDataFlagOn();
} 
  

/*----------------------------------------------------------------------*/
template <class TImage>
ImageToPolyDataFilter<TImage>::~ImageToPolyDataFilter()
{
  if( m_ContourFilter )
    m_ContourFilter->Delete();

  if( m_PolyDataNormals )
    m_PolyDataNormals->Delete();

  if( m_GaussianFilter ) 
    m_GaussianFilter->Delete();

  if( m_ImageThreshold )
    m_ImageThreshold->Delete();

  if(m_DecimatePro)
    m_DecimatePro->Delete();
}


/*----------------------------------------------------------------------*/
template <class TImage>
void ImageToPolyDataFilter<TImage>::SetInput( TImage * image ) 
{
  m_Input = TImage::New();
  m_Input = image;
}


/*----------------------------------------------------------------------*/
template <class TImage>
vtkPolyData* ImageToPolyDataFilter<TImage>::GetOutput()
{
  return m_PolyDataNormals->GetOutput();
}


/*----------------------------------------------------------------------*/
template <class TImage>
void ImageToPolyDataFilter<TImage>::ThreadedUpdate()
{
  // Call the superclass implementation to spawn the thread
  //Superclass::ThreadedUpdate();
}

/*----------------------------------------------------------------------*/
template <class TImage>
void ImageToPolyDataFilter<TImage>::SetDecimationTarget( float target )
{
  if( m_DecimatePro )
    m_DecimatePro->SetTargetReduction(target);
}

/*----------------------------------------------------------------------*/
template <class TImage>
void ImageToPolyDataFilter<TImage>::Update()
{

  this->ConnectPipeline();

  this->SetProgress( 0.0f );

  float gaussianWeight, contourWeight, decimateWeight;

  if( m_GaussianFilter )
  {
    gaussianWeight = 0.3f;
    contourWeight  = 0.4f;
    decimateWeight = 0.1f;
  }
  else
  {
    gaussianWeight = 0.0f;
    contourWeight  = 0.5f;
    decimateWeight = 0.2f;
  }
    
  if( m_GaussianFilter )
  { 
    m_GaussianFilter->Update();    
  } 
  m_ContourFilter->Update();
  
 
  if( m_MaxPolygons ) // a max number of polys was set
  {
    double polys = static_cast<double>( m_ContourFilter->GetOutput()->GetNumberOfPolys() );
    if( polys > m_MaxPolygons )
      m_DecimatePro->SetTargetReduction( 1.0 - ( static_cast<double>( m_MaxPolygons ) / polys ) );
  }
  m_DecimatePro->Update();

  m_PolyDataNormals->Update();

}


/*----------------------------------------------------------------------*/
template <class TImage>
vtkDataSet * ImageToPolyDataFilter<TImage>::GetVtkImage()
{
  return m_ITKToVTKAdapter->GetOutput();
}


/*----------------------------------------------------------------------*/
template <class TImage>
void ImageToPolyDataFilter<TImage>::ConnectPipeline()  
{

  // generate the initial pipeline 
  m_ITKToVTKAdapter->SetInput( m_Input ); //set the input
  m_ContourFilter->SetValue( 0, m_IsoValue );   //set the iso surface value

  if( m_SmoothingSigma )  //if smoothing enabled
  {
    if( !m_GaussianFilter )
    {
      m_GaussianFilter = vtkImageGaussianSmooth::New();
      m_GaussianFilter->ReleaseDataFlagOn();
    }

    // If the image is a mask, scale it to be able to apply smoothing
    if( m_InputIsMaskImage )  
    {
      if( !m_ImageThreshold )
      {
        m_ImageThreshold = vtkImageThreshold::New();
        m_ImageThreshold->ReleaseDataFlagOn();
      }

      m_ImageThreshold->ThresholdBetween( m_IsoValue, m_IsoValue );
      m_ImageThreshold->SetOutputScalarTypeToUnsignedChar();
      m_ImageThreshold->ReplaceInOn();
      m_ImageThreshold->SetInValue(255);
#if VTK_MAJOR_VERSION <= 5
      m_ImageThreshold->SetInput( m_ITKToVTKAdapter->GetOutput() );
#else
      m_ImageThreshold->SetInputData( m_ITKToVTKAdapter->GetOutput() );
#endif 

#if VTK_MAJOR_VERSION <= 5
      m_GaussianFilter->SetInput( m_ImageThreshold->GetOutput() );
#else
      m_GaussianFilter->SetInputData( m_ImageThreshold->GetOutput() );
#endif 

      m_ContourFilter->SetValue( 0, 127.5 );  //set the value to the middle one     
    }
    else  //connect directly to the gaussian filter
#if VTK_MAJOR_VERSION <= 5
			m_GaussianFilter->SetInput( m_ITKToVTKAdapter->GetOutput() );
#else
			m_GaussianFilter->SetInputData( m_ITKToVTKAdapter->GetOutput() );
#endif      

    m_GaussianFilter->SetStandardDeviation( m_SmoothingSigma );

#if VTK_MAJOR_VERSION <= 5
    m_ContourFilter->SetInput( m_GaussianFilter->GetOutput() );
#else
    m_ContourFilter->SetInputData( m_GaussianFilter->GetOutput() );
#endif
  }
  else
#if VTK_MAJOR_VERSION <= 5
    m_ContourFilter->SetInput( m_ITKToVTKAdapter->GetOutput() );
#else
    m_ContourFilter->SetInputData( m_ITKToVTKAdapter->GetOutput() );
#endif

#if VTK_MAJOR_VERSION <= 5
  m_DecimatePro->SetInput( m_ContourFilter->GetOutput() );
#else
  m_DecimatePro->SetInputData( m_ContourFilter->GetOutput() );
#endif
  m_DecimatePro->SetFeatureAngle( m_FeatureAngle );

#if VTK_MAJOR_VERSION <= 5
  m_PolyDataNormals->SetInput( m_DecimatePro->GetOutput() );
#else
  m_PolyDataNormals->SetInputData( m_DecimatePro->GetOutput() );
#endif

}


#endif // ImageToPolyDataFilter_txx
