/*=======================================================================

  Program:   
  Module:    $RCSfile: $
  Language:  C++
  Date:      $Date: $
  Version:   $Revision: $
  Author:    $Author: $
  Comment:   

=========================================================================*/



// Project-specific includes
#include "ImageToPolyDataFilter.h"

// ITK includes
#include "itkImage.h"
#include "itkFlipImageFilter.h"

// VTK includes
#include <vtkActor.h>
#include <vtkDataSetMapper.h>
#include <vtkClipPolyData.h>
#include <vtkCylinder.h>
#include <vtkCommand.h>
#include <vtkExtractPolyDataGeometry.h>
#include <vtkInteractorStyleTrackballCamera.h>
#include <vtkOutlineFilter.h>
#include <vtkPolyDataMapper.h>
#include <vtkProperty.h>
#include <vtkRenderer.h>
#include <vtkRenderWindow.h>
#include <vtkRenderWindowInteractor.h>
#include <vtkSampleFunction.h>
#include <vtkSmartPointer.h>
#include <vtkSphere.h>
#include <vtkTransform.h>



 /** 
 *  \brief This program shows how to crop an input dataset with an implicit function.
 *  
 *  This program performs cropping actions defined by the boundaries of an implicit function
 *  over an input dataset, which is expected to be a segmentation mask. The mask is read 
 *  from disk as an ITK image, and then converted to a surface, represented by a 
 *  vtkPolyData.
 *
 *	\param inputImageFile Name (including path) of the input dataset (*.mhd/*.mha) file
 *	\param isoValue Iso-value to create the isosurface with VTK
 * 
 *	\return zero if successful; non-zero otherwise
 *
 *  
 */ 
 
int main(int argc, char* argv[])
{

  if(argc < 2 )
    {
    std::cerr << "Usage: " << std::endl;
    std::cerr << argv[0] << " inputImageFile [isoValue]" << std::endl;
    return EXIT_FAILURE;
    }

  typedef unsigned char                            PixelType;
  typedef itk::Image<PixelType,3>                  ImageType;
      
  typedef itk::ImageFileReader<ImageType>     ReaderType;

  typedef itk::FlipImageFilter<ImageType>      ITKFlipImageFilterType;      
  ITKFlipImageFilterType::Pointer flipFilter = ITKFlipImageFilterType::New();

  typedef ImageToPolyDataFilter<ImageType>       ImageToPolyDataFilterType;  
  ImageToPolyDataFilterType::Pointer imageToPolyDataFilter = ImageToPolyDataFilterType::New();


  ReaderType::Pointer	reader = ReaderType::New();  
  
  // Read the input dataset
  reader->SetFileName(argv[1]);

#ifdef _DEBUG
  std::cout << "Rading the input image... ";
#endif

  try
  {
    reader->Update();
  }
  catch ( itk::ExceptionObject &excp )
  {
    std::cerr << "Exception thrown while reading the image" << std::endl;
    std::cerr << excp << std::endl;        
    return EXIT_FAILURE;
  }

#ifdef _DEBUG
  std::cout << "Done!" << std::endl;
#endif

  flipFilter->SetInput( reader->GetOutput() );

  itk::FixedArray<bool,3> flipAxes;
  flipAxes[0] = false;
  flipAxes[1] = true;
  flipAxes[2] = false;

  flipFilter->SetFlipAxes( flipAxes );  

  // Need to flip the itk image to align vtk and itk coordinates

#ifdef _DEBUG
  std::cout << "Flipping the axes... ";
#endif

  try
  {
    flipFilter->Update();
  }
  catch ( itk::ExceptionObject &excp )
  {
    std::cerr << "Exception thrown while flipping the image" << std::endl;
    std::cerr << excp << std::endl;        
    return EXIT_FAILURE;
  }

#ifdef _DEBUG
  std::cout << "Done!" << std::endl;
#endif

  // Convert the VTK image to a VTK PolyData
  double isoValue = 127.5;
  if(argc > 2 )
    isoValue = atof(argv[2]);

  imageToPolyDataFilter->SetInput(flipFilter->GetOutput());
  imageToPolyDataFilter->SetIsoValue(isoValue);

#ifdef _DEBUG
  std::cout << "Converting the ITK image to a VTK polydata... ";
#endif

  try {
   imageToPolyDataFilter->Update();
  }
  catch ( itk::ExceptionObject &excp )
  {
    std::cerr << "Exception thrown while converting the ITK image to a VTK polydata" << std::endl;
    std::cerr << excp << std::endl;        
    return EXIT_FAILURE;
  }

#ifdef _DEBUG
  std::cout << "Done!" << std::endl;
#endif

  vtkSmartPointer<vtkPolyData> polyData = vtkSmartPointer<vtkPolyData>::New();
  polyData->ShallowCopy(imageToPolyDataFilter->GetOutput()); 
   

  // Define a transform for the implicit function
  vtkSmartPointer<vtkTransform> t = 
    vtkSmartPointer<vtkTransform>::New();
  t->RotateX(90);

  // Define the implicit function as a cylinder: remember, the cylinder is 
  // infinite in extent. To truncate the cylinder use the vtkImplicitBoolean 
  // in combination with clipping planes.
  vtkSmartPointer<vtkCylinder> cylinderImplFunc = 
    vtkSmartPointer<vtkCylinder>::New();
  cylinderImplFunc->SetRadius(20.25);
  cylinderImplFunc->SetTransform(t);

  // Extract the vtkPolyData cells that lie either entirely inside the specified 
  // implicit function 
  vtkSmartPointer<vtkExtractPolyDataGeometry> extract = 
    vtkSmartPointer<vtkExtractPolyDataGeometry>::New();
  extract->SetInput(polyData);
  extract->SetImplicitFunction(cylinderImplFunc);
  extract->ExtractBoundaryCellsOn();
#if VTK_MAJOR_VERSION > 5
  extract->PassPointsOn();
#endif

 

  // Sample the function
  vtkSmartPointer<vtkSampleFunction> sample = 
    vtkSmartPointer<vtkSampleFunction>::New();
  sample->SetSampleDimensions(50,50,50);
  sample->SetImplicitFunction(cylinderImplFunc);
  double value = 100.0;
  double xmin = -value, xmax = value,
    ymin = -value, ymax = value,
    zmin = -value, zmax = value;
  sample->SetModelBounds(xmin, xmax, ymin, ymax, zmin, zmax);
 
  // Create the 0 isosurface
  vtkSmartPointer<vtkContourFilter> contours = 
    vtkSmartPointer<vtkContourFilter>::New();
  contours->SetInputConnection(sample->GetOutputPort());
  contours->GenerateValues(1, 1, 1);



  // Create rendering mappers and actors
  vtkSmartPointer<vtkPolyDataMapper> polyDataMapper 
    = vtkSmartPointer<vtkPolyDataMapper>::New();
#if VTK_MAJOR_VERSION <= 5
  polyDataMapper->SetInput(polyData);
#else
  polyDataMapper->SetInputData(polyData);
#endif
  polyDataMapper->ScalarVisibilityOff();
  polyDataMapper->Update();

  vtkSmartPointer<vtkActor> polyDataActor = vtkSmartPointer<vtkActor>::New();
  polyDataActor->SetMapper(polyDataMapper);
  polyDataActor->GetProperty()->BackfaceCullingOff();
  polyDataActor->GetProperty()->SetColor(255.0/255.0, 255.0/255.0, 255.0/255.0); // white


  // Map the contours to graphical primitives
  vtkSmartPointer<vtkPolyDataMapper> contourMapper = 
    vtkSmartPointer<vtkPolyDataMapper>::New();
  contourMapper->SetInputConnection(contours->GetOutputPort());
  contourMapper->ScalarVisibilityOff();
  //contourMapper->SetScalarRange(0.0, 1.2);

  vtkSmartPointer<vtkActor> contourActor = 
    vtkSmartPointer<vtkActor>::New();
  contourActor->SetMapper(contourMapper);
  contourActor->GetProperty()->SetColor(255.0/255.0, 0.0/255.0, 0.0/255.0); // red


  vtkSmartPointer<vtkPolyDataMapper> extractMapper = 
    vtkSmartPointer<vtkPolyDataMapper>::New();
  extractMapper->SetInputConnection(extract->GetOutputPort());
  extractMapper->GlobalImmediateModeRenderingOn();

  extractMapper->ScalarVisibilityOff();
  extractMapper->Update();

  vtkSmartPointer<vtkActor> extractActor = vtkSmartPointer<vtkActor>::New();
  extractActor->SetMapper(extractMapper);
  extractActor->GetProperty()->BackfaceCullingOff();
  extractActor->GetProperty()->SetColor(255.0/255.0, 255.0/255.0, 255.0/255.0); // white


  // Define viewport ranges
  // (xmin, ymin, xmax, ymax)
  double numberOfViewports = 3.;
  double Viewport1[4] = {0.0/numberOfViewports, 0.0, 1.0/numberOfViewports, 1.0};
  double Viewport2[4] = {1.0/numberOfViewports, 0.0, 2.0/numberOfViewports, 1.0};
  double Viewport3[4] = {2.0/numberOfViewports, 0.0, 3.0/numberOfViewports, 1.0};

  // Define graphics objects
  // The first render window will display only the input dataset as a polyData
  vtkSmartPointer<vtkRenderer> renderer1 = vtkSmartPointer<vtkRenderer>::New();
  renderer1->SetViewport(Viewport1);
  renderer1->SetBackground(0.0, 0.0, 0.0); // black
  renderer1->AddActor(polyDataActor);

  // The second render window will display the implicit function
  vtkSmartPointer<vtkRenderer> renderer2 = vtkSmartPointer<vtkRenderer>::New();
  renderer2->SetViewport(Viewport2);
  renderer2->SetBackground(0.0, 0.0, 0.0); // black
  renderer2->AddActor(contourActor);

  // The second render window will display the extracted polyData
  vtkSmartPointer<vtkRenderer> renderer3 = vtkSmartPointer<vtkRenderer>::New();
  renderer3->SetViewport(Viewport3);
  renderer3->SetBackground(0.0, 0.0, 0.0); // black
  renderer3->AddActor(extractActor);

  vtkSmartPointer<vtkRenderWindow> renWin = 
    vtkSmartPointer<vtkRenderWindow>::New();

  //renWin->SetMultiSamples(0);
  renWin->SetSize(300*numberOfViewports,300); //(width, height)

  renWin->AddRenderer(renderer1);
  renWin->AddRenderer(renderer2);
  renWin->AddRenderer(renderer3);


  vtkSmartPointer<vtkRenderWindowInteractor> iren = 
    vtkSmartPointer<vtkRenderWindowInteractor>::New();
  iren->SetRenderWindow(renWin);
 
  vtkSmartPointer<vtkInteractorStyleTrackballCamera> style = 
    vtkSmartPointer<vtkInteractorStyleTrackballCamera>::New();
  iren->SetInteractorStyle(style);

  iren->Initialize();
  iren->Start();

 
  return EXIT_SUCCESS;
}
