/*=======================================================================

  Program:   
  Module:    $RCSfile: $
  Language:  C++
  Date:      $Date: $
  Version:   $Revision: $
  Author:    $Author: $
  Comment:   

=========================================================================*/



// Project-specific includes
#include "ImageToPolyDataFilter.h"

// ITK includes
#include "itkImage.h"
#include "itkFlipImageFilter.h"

// VTK includes
#include <vtkActor.h>
#include <vtkBoxWidget.h>
#include <vtkDataSetMapper.h>
#include <vtkClipPolyData.h>
#include <vtkCylinder.h>
#include <vtkCommand.h>
#include <vtkExtractPolyDataGeometry.h>
#include <vtkInteractorStyleTrackballCamera.h>
#include <vtkOutlineFilter.h>
#include <vtkPlanes.h>
#include <vtkPolyDataMapper.h>
#include <vtkProperty.h>
#include <vtkRenderer.h>
#include <vtkRenderWindow.h>
#include <vtkRenderWindowInteractor.h>
#include <vtkSmartPointer.h>
#include <vtkSphere.h>
#include <vtkSphereSource.h>
#include <vtkTransform.h>



 /** 
 *  \brief This program shows how to crop an input dataset with a vtkBoxWidget.
 *  
 *  This program performs cropping actions defined by the boundaries of a vtkBoxWidget
 *  over an input dataset, which is expected to be a segmentation mask. The mask is read 
 *  from disk as an ITK image, and then converted to a surface, represented by a 
 *  vtkPolyData.
 *
 *	\param inputImageFile Name (including path) of the input dataset (*.mhd/*.mha) file
 *	\param isoValue Iso-value to create the isosurface with VTK
 * 
 *	\return zero if successful; non-zero otherwise
 *
 */ 


class vtkMyCallback : public vtkCommand
{
public:
  
  vtkPlanes* planes;
  vtkActor* inputActor;
  vtkSmartPointer<vtkActor> extractActor;
  vtkSmartPointer<vtkRenderer> cropRenderer;

  static vtkMyCallback *New() 
    { return new vtkMyCallback; }
  virtual void Execute(vtkObject *caller, unsigned long, void*)
    {
      vtkSmartPointer<vtkTransform> t = 
	      vtkSmartPointer<vtkTransform>::New();
      vtkBoxWidget *widget = reinterpret_cast<vtkBoxWidget*>(caller);
      widget->GetTransform(t);

      vtkSmartPointer<vtkPropCollection> propCollection = 
	      vtkSmartPointer<vtkPropCollection>::New();
      widget->GetProp3D()->GetActors(propCollection);
      
      inputActor = dynamic_cast<vtkActor *>(propCollection->GetLastProp());

      planes = vtkPlanes::New();
      vtkSmartPointer<vtkPolyDataMapper> polyDataMapper 
        = vtkSmartPointer<vtkPolyDataMapper>::New();

      widget->GetPlanes(planes);
      //inputActor->GetMapper()->SetClippingPlanes(planes);

      this->CropWithImplicitFunction();

    }
    void SetCropRenderer(vtkRenderer *renderer)
    {
      cropRenderer = renderer;
    }

    void CropWithImplicitFunction()
    {
      // This portion of the code clips the mace with the vtkPlanes implicit
      // function. The clipped region is colored green.

      vtkSmartPointer<vtkExtractPolyDataGeometry> extractor = 
        vtkSmartPointer<vtkExtractPolyDataGeometry>::New();

      extractor->SetInput(inputActor->GetMapper()->GetInput());
      extractor->SetImplicitFunction(planes);
      extractor->ExtractBoundaryCellsOn();
#if VTK_MAJOR_VERSION > 5
      extractor->PassPointsOff();
#endif
      vtkSmartPointer<vtkPolyDataMapper> extractMapper = 
        vtkSmartPointer<vtkPolyDataMapper>::New();

      extractMapper->SetInputConnection(extractor->GetOutputPort());

      extractActor = vtkSmartPointer<vtkActor>::New();
      extractActor->SetMapper(extractMapper);  
      extractActor->GetProperty()->SetColor(1.0, 0.0, 0.0);

      // Define graphics objects
      cropRenderer->SetBackground(0.0, 0.0, 0.0); // black
      cropRenderer->AddActor(extractActor);
    }
};

 
int main(int argc, char* argv[])
{

  if(argc < 2 )
    {
    std::cerr << "Usage: " << std::endl;
    std::cerr << argv[0] << " inputImageFile [isoValue]" << std::endl;
    return EXIT_FAILURE;
    }

  typedef unsigned char                            PixelType;
  typedef itk::Image<PixelType,3>                  ImageType;
      
  typedef itk::ImageFileReader<ImageType>     ReaderType;

  typedef itk::FlipImageFilter<ImageType>      ITKFlipImageFilterType;      
  ITKFlipImageFilterType::Pointer flipFilter = ITKFlipImageFilterType::New();

  typedef ImageToPolyDataFilter<ImageType>       ImageToPolyDataFilterType;  
  ImageToPolyDataFilterType::Pointer imageToPolyDataFilter = ImageToPolyDataFilterType::New();


  ReaderType::Pointer	reader = ReaderType::New();  
  
  // Read the input dataset
  reader->SetFileName(argv[1]);

#ifdef _DEBUG
  std::cout << "Rading the input image... ";
#endif

  try
  {
    reader->Update();
  }
  catch ( itk::ExceptionObject &excp )
  {
    std::cerr << "Exception thrown while reading the image" << std::endl;
    std::cerr << excp << std::endl;        
    return EXIT_FAILURE;
  }

#ifdef _DEBUG
  std::cout << "Done!" << std::endl;
#endif

  flipFilter->SetInput( reader->GetOutput() );

  itk::FixedArray<bool,3> flipAxes;
  flipAxes[0] = false;
  flipAxes[1] = true;
  flipAxes[2] = false;

  flipFilter->SetFlipAxes( flipAxes );  

  // Need to flip the itk image to align vtk and itk coordinates

#ifdef _DEBUG
  std::cout << "Flipping the axes... ";
#endif

  try
  {
    flipFilter->Update();
  }
  catch ( itk::ExceptionObject &excp )
  {
    std::cerr << "Exception thrown while flipping the image" << std::endl;
    std::cerr << excp << std::endl;        
    return EXIT_FAILURE;
  }

#ifdef _DEBUG
  std::cout << "Done!" << std::endl;
#endif

  // Convert the VTK image to a VTK PolyData
  double isoValue = 127.5;
  if(argc > 2 )
    isoValue = atof(argv[2]);

  imageToPolyDataFilter->SetInput(flipFilter->GetOutput());
  imageToPolyDataFilter->SetIsoValue(isoValue);

#ifdef _DEBUG
  std::cout << "Converting the ITK image to a VTK polydata... ";
#endif

  try {
   imageToPolyDataFilter->Update();
  }
  catch ( itk::ExceptionObject &excp )
  {
    std::cerr << "Exception thrown while converting the ITK image to a VTK polydata" << std::endl;
    std::cerr << excp << std::endl;        
    return EXIT_FAILURE;
  }

#ifdef _DEBUG
  std::cout << "Done!" << std::endl;
#endif

  vtkSmartPointer<vtkPolyData> polyData = vtkSmartPointer<vtkPolyData>::New();
  polyData->ShallowCopy(imageToPolyDataFilter->GetOutput()); 
   
  vtkSmartPointer<vtkPolyDataMapper> polyDataMapper 
    = vtkSmartPointer<vtkPolyDataMapper>::New();
#if VTK_MAJOR_VERSION <= 5
  polyDataMapper->SetInput(polyData);
#else
  polyDataMapper->SetInputData(polyData);
#endif
  polyDataMapper->ScalarVisibilityOff();
  polyDataMapper->Update();

  vtkSmartPointer<vtkActor> polyDataActor = vtkSmartPointer<vtkActor>::New();
  polyDataActor->SetMapper(polyDataMapper);
  polyDataActor->GetProperty()->BackfaceCullingOff();
  polyDataActor->GetProperty()->SetColor(255.0/255.0, 255.0/255.0, 255.0/255.0); // white


  // Define viewport ranges
  // (xmin, ymin, xmax, ymax)
  double numberOfViewports = 2.;
  double Viewport1[4] = {0.0/numberOfViewports, 0.0, 1.0/numberOfViewports, 1.0};
  double Viewport2[4] = {1.0/numberOfViewports, 0.0, 2.0/numberOfViewports, 1.0};


  // Define graphics objects
  vtkSmartPointer<vtkRenderer> renderer = vtkSmartPointer<vtkRenderer>::New();
  renderer->SetViewport(Viewport1);
  renderer->SetBackground(0.0, 0.0, 0.0); // black
  renderer->AddActor(polyDataActor);


  vtkSmartPointer<vtkRenderer> cropRenderer = vtkSmartPointer<vtkRenderer>::New();
  cropRenderer->SetViewport(Viewport2);
  cropRenderer->SetBackground(0.0, 0.0, 0.0); // black


  vtkSmartPointer<vtkRenderWindow> renWin = vtkSmartPointer<vtkRenderWindow>::New();
  //renWin->SetMultiSamples(0);
  renWin->AddRenderer(renderer);
  renWin->AddRenderer(cropRenderer);

  vtkSmartPointer<vtkRenderWindowInteractor> iren = vtkSmartPointer<vtkRenderWindowInteractor>::New();
  iren->SetRenderWindow(renWin);
 
  vtkSmartPointer<vtkInteractorStyleTrackballCamera> style = 
    vtkSmartPointer<vtkInteractorStyleTrackballCamera>::New();
  iren->SetInteractorStyle(style);


  // Add a box widget
  vtkSmartPointer<vtkBoxWidget> boxWidget = vtkSmartPointer<vtkBoxWidget>::New();
  boxWidget->SetInteractor(iren);
  boxWidget->SetPlaceFactor(1.0);
  boxWidget->SetProp3D(polyDataActor);
  boxWidget->PlaceWidget();
  vtkSmartPointer<vtkMyCallback> callback =
    vtkSmartPointer<vtkMyCallback>::New();

  callback->SetCropRenderer(cropRenderer);

  boxWidget->AddObserver(vtkCommand::InteractionEvent, callback);
 
  boxWidget->On();

  // Switch handles (the spheres) on
  boxWidget->HandlesOn();

  // Set the normal inside out (otherwise the volume would not be displayed)
  boxWidget->InsideOutOn();

  boxWidget->RotationEnabledOff();

  //renWin->Render();

  iren->Initialize();
  iren->Start();
 
  return EXIT_SUCCESS;
}
